decode = require './decode'
ord    = require './ord'
chr    = require './chr'
require 'buffertools'

module.exports.ProtocolNode = class ProtocolNode
  constructor: (@tag, @attributeHash, @children=[], @data) ->
    # console.log "[#{@tag}] NEW NODE WITH CHILDREN", @children.length

  NodeString: (indent='') ->
    ret = "\n#{indent}<#{@tag}"
    if @attributeHash?
      for key, value of @attributeHash
        ret += " #{key}=\"#{value}\""

    ret += ">"

    if @data.length > 0
      ret += @data

    if @children.length > 0
      for child in @children
        ret += child.NodeString "#{indent}  "
      ret += "\n#{indent}"

    ret += "</#{@tag}>"
    return ret

  getAttribute: (attribute) ->
    ret = ""
    if @attributeHash.hasOwnProperty attribute
      ret = @attributeHash[attribute]
    return ret

  getChild: (tag) ->
    ret = null
    if @children.length > 0
      for child in @children
        if child.tag is tag
          return child

        ret = child.getChild(tag)
        if ret
          return ret

    return null

module.exports.BinTreeNodeReader = class BinTreeNodeReader
  constructor: (@dictionary) ->

  nextTree: (input) ->
    if input?
      @input = input

    stanzaSize = @peekInt24()
    flags = (stanzaSize >> 20)
    size = ((stanzaSize & 0xF0000) >> 16) | ((stanzaSize & 0xFF00) >> 8) | (stanzaSize & 0xFF)

    isEncrypted = (flags & 8) isnt 0

    if size > @input.length
      exception = new IncompleteMessageException 'Incomplete message'
      exception.setInput @input
      throw exception

    @readInt24()
    if size > 0
      if isEncrypted and @encryptionKey isnt null
        console.log "should read encrypted tree here..."
        return
      return @nextTreeInternal()

    return null

  getToken: (token) ->
    ret = ""
    if token >= 0 and token < Object.keys(@dictionary).length
      ret = @dictionary[token]
    else
      return console.log "BinTreeNodeReader.getToken(): Invalid token #{token}"

  readString: (token) ->
    ret = ""
    if token is -1
      return console.log "BinTreeNodeReader.readString(): Invalid token #{token}"

    if token > 4 and token < 0xf5
      ret = @getToken token
    else if token is 0
      ret = ""
    else if token is 0xfc
      size = @readInt8()
      ret = @fillArray size
    else if token is 0xfd
      size = @readInt24()
      ret = @fillArray size
    else if token is 0xfe
      size = @readInt8()
      ret = @getToken token + 0xf5
    else if token is 0xfa
      user = @readString @readInt8()
      server = @readString @readInt8()
      if user.length > 0 and server.length > 0
        ret = "#{user}@#{server}"
      else if server.length is 0
        ret = server

    return ret

  readAttributes: (size) ->
    attributes = []
    attribCount = (size - 2 + size % 2) / 2
    for i in [0...attribCount]
      key = @readString @readInt8()
      value = @readString @readInt8()
      attributes[key] = value
    return attributes

  nextTreeInternal: ->
    token = @readInt8()
    size = @readListSize token
    token = @readInt8()

    if token is 1
      attributes = @readAttributes size
      return new ProtocolNode "start", attributes, null, ""
    else if token is 2
      return null

    tag = @readString token
    attributes = @readAttributes size
    if size % 2 is 1
      return new ProtocolNode tag, attributes, null, ""

    token = @readInt8()
    if @isListTag token
      return new ProtocolNode tag, attributes, @readList(token), ""

    return new ProtocolNode tag, attributes, null, @readString(token)

  isListTag: (token) ->
    return ((token is 248) or (token is 0) or (token is 249))

  readList: (token) ->
    size = @readListSize token
    ret = []
    for i in [0...size]
      ret.push @nextTreeInternal()
    return ret

  readListSize: (token) ->
    size = 0
    if token is 0xf8
      size = @readInt8()
    else if token is 0xf9
      size = @readInt16()
    else
      throw new Error("BinTreeNodeReader.readListSize(): Invalid token #{token}")
    return size


  readInt24: ->
    ret = @peekInt24()
    if ret > 0
      @input = @input.slice 3
    return ret

  peekInt24: ->
    ret = 0
    if @input.length >= 3
      ret = @input[0] << 16
      ret |= @input[1] << 8
      ret |= @input[2] << 0
    return ret

  peekInt16: ->
    ret = 0
    if @input.length >= 2
      ret = @input[0] << 8
      ret |= @input[1] << 0
    return ret

  readInt16: ->
    ret = @peekInt16()
    if ret > 0
      @input = @input.slice 2
    return ret

  readInt8: ->
    ret = 0
    if @input.length >= 1
      ret = @input[0]
      @input = @input.slice 1
    return ret

  fillArray: (len) ->
    ret = new Buffer(0)
    if @input.length >= len
      ret = @input.slice 0, len
      @input = @input.slice len
  
    return ret

module.exports.BinTreeNodeWriter = class BinTreeNodeWriter
  constructor: (dictionary) ->
    @tokenMap = []
    @output = null;

    for i in [0...Object.keys(dictionary).length]
      if dictionary[i].length > 0
        @tokenMap[dictionary[i]] = i

  startStream: (domain, resource) ->
    @output = [ord('W'), ord('A'), 0x01, 0x02, 0x00, 0x00, 0x19]

    attributes = []
    attributes.to = domain
    attributes.resource = resource

    @writeListStart Object.keys(attributes).length * 2 + 1

    @output.push 0x01
    @writeAttributes attributes

    ret = @output

    @output = []

    return ret

  write: (node) ->
    if node is null
      @output.push 0x00
    else
      @writeInternal node
    return @flushBuffer()

  writeInternal: (node) ->
    len = 1
    if node.attributeHash isnt null
      len += Object.keys(node.attributeHash).length * 2
    if node.children.length > 0
      len += 1
    if node.data.length > 0
      len += 1

    @writeListStart len
    @writeString node.tag
    @writeAttributes node.attributeHash

    if node.data.length > 0
      @writeBytes node.data

    if node.children.length > 0
      @writeListStart node.children.length
      for child in node.children
        @writeInternal child

  flushBuffer: ->
    size = @output.length
    ret = @getInt24(size).concat(@output)

    @output = []
    return ret

  writeToken: (token) ->
    if token < 0xf5
      @output.push token
    else if token <= 0x1f4
      @output = @output.concat [0xfe, token - 0xf5]

  writeJid: (user, server) ->
    @output.push 0xfa
    if user.length > 0
      @writeString user
    else
      @writeToken 0
    @writeString server

  writeInt8: (v) ->
    @output.push v & 0xff

  writeInt16: (v) ->
    ret = (v & 0xff00) >> 8
    ret += (v & 0x00ff) >> 0
    return ret

  getInt24: (v) ->
    ret =    [(v & 0xff0000) >> 1]
    ret.push (v & 0x00ff00) >> 8
    ret.push (v & 0x0000ff) >> 0

    return ret

  writeInt24: (v) ->
    @output.push (v & 0xff0000) >> 1
    @output.push (v & 0x00ff00) >> 8
    @output.push (v & 0x0000ff) >> 0

  writeBytes: (bytes) ->
    if bytes instanceof Array
      len = bytes.length
    else
      len = bytes.toString().length

    if len >= 0x100
      @output.push 0xfd
      @writeInt24 len
    else
      @output.push 0xfc
      @writeInt8 len

    if bytes instanceof Array
      @output = @output.concat(bytes)
    else
      for byte in bytes.split("")
        @output.push ord(byte)

  writeString: (tag) ->
    if @tokenMap.hasOwnProperty tag
      key = @tokenMap[tag]
      @writeToken key
    else
      index = tag.toString().indexOf '@'
      if index >= 0
        server = tag.substr(index + 1)
        user   = tag.substr(0, index)
        @writeJid user, server
      else
        @writeBytes tag

  writeAttributes: (attributes) ->
    if attributes
      for key, value of attributes
        @writeString key
        @writeString value

  writeListStart: (len) ->
    if len is 0
      @output.push 0x00
    else if len < 256
      @output = @output.concat [0xf8, len]
    else
      @output = @output.concat [0xf9, len]